import asyncio
import socket
import time
from functools import partial
from math import floor
from typing import Callable

import psutil

from boneio.const import (
    CPU,
    DISK,
    GIGABYTE,
    HOST,
    MEGABYTE,
    MEMORY,
    NETWORK,
    SWAP,
    UPTIME,
    IP,
    MASK,
    MAC,
    NONE,
)
from boneio.sensor import LM75Sensor, MCP9808Sensor
from boneio.version import __version__


intervals = (("d", 86400), ("h", 3600), ("m", 60))


def display_time(seconds):
    """Strf time."""
    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip("s")
            result.append(f"{int(value)}{name}")
    return "".join(result)


async def get_network_info(host_data):
    """Fetch network info."""

    def retrieve_from_psutil():
        addrs = psutil.net_if_addrs()["eth0"]
        out = {IP: NONE, MASK: NONE, MAC: NONE}
        for addr in addrs:
            if addr.family == socket.AF_INET:
                out["ip"] = addr.address
                out["mask"] = addr.netmask
            elif addr.family == psutil.AF_LINK:
                out["mac"] = addr.address
        return out

    while True:
        data = retrieve_from_psutil()
        host_data.write(NETWORK, data)
        await asyncio.sleep(60)


async def get_cpu_info(host_data):
    """Fetch CPU info."""
    while True:
        cpu = psutil.cpu_times_percent()
        host_data.write(
            CPU,
            {
                "total": f"{int(100 - cpu.idle)}%",
                "user": f"{cpu.user}%",
                "system": f"{cpu.system}%",
            },
        )
        await asyncio.sleep(5)


async def get_disk_info(host_data):
    """Fetch disk info."""
    while True:
        disk = psutil.disk_usage("/")
        host_data.write(
            DISK,
            {
                "total": f"{floor(disk.total / GIGABYTE)}GB",
                "used": f"{floor(disk.used / GIGABYTE)}GB",
                "free": f"{floor(disk.free / GIGABYTE)}GB",
            },
        )
        await asyncio.sleep(60)


async def get_memory_info(host_data):
    """Fetch memory info."""
    while True:
        vm = psutil.virtual_memory()
        host_data.write(
            MEMORY,
            {
                "total": f"{floor(vm.total / MEGABYTE)}MB",
                "used": f"{floor(vm.used / MEGABYTE)}MB",
                "free": f"{floor(vm.available / MEGABYTE)}MB",
            },
        )
        await asyncio.sleep(10)


async def get_swap_info(host_data):
    """Fetch swap info."""
    while True:
        swap = psutil.swap_memory()
        host_data.write(
            SWAP,
            {
                "total": f"{floor(swap.total / MEGABYTE)}MB",
                "used": f"{floor(swap.used / MEGABYTE)}MB",
                "free": f"{floor(swap.free / MEGABYTE)}MB",
            },
        )
        await asyncio.sleep(10)


async def get_uptime(host_data):
    """Fetch uptime info."""
    while True:
        uptime = display_time(time.clock_gettime(time.CLOCK_BOOTTIME))
        host_data.write_uptime(uptime)
        await asyncio.sleep(30)


host_stats = {
    NETWORK: get_network_info,
    CPU: get_cpu_info,
    DISK: get_disk_info,
    MEMORY: get_memory_info,
    SWAP: get_swap_info,
    UPTIME: get_uptime,
}


class HostData:
    """Helper class to store host data."""

    data = {UPTIME: {}, NETWORK: {}, CPU: {}, DISK: {}, MEMORY: {}, SWAP: {}}

    def __init__(
        self,
        output: dict,
        callback: Callable,
        temp_sensor: Callable[[LM75Sensor, MCP9808Sensor], None],
    ) -> None:
        """Initialize HostData."""
        self._hostname = socket.gethostname()
        self.data[UPTIME] = {HOST: self._hostname, UPTIME: 0, "version": __version__}
        self._temp_sensor = temp_sensor
        self._output = output
        self._callback = callback
        self._loop = asyncio.get_running_loop()

    def write(self, type: str, data: dict) -> None:
        """Write data of chosen type."""
        self.data[type] = data
        self._loop.call_soon_threadsafe(partial(self._callback, type))

    def write_uptime(self, uptime: str) -> None:
        """Write uptime."""
        self.data[UPTIME][UPTIME] = uptime
        if self._temp_sensor:
            self.data[UPTIME][self._temp_sensor.name] = self._temp_sensor.state
        self._loop.call_soon_threadsafe(partial(self._callback, UPTIME))

    def get(self, type: str) -> dict:
        """Get saved stats."""
        if type in self._output:
            return self._get_output(type)
        return self.data[type]

    def _get_output(self, type: str) -> dict:
        """Get stats for output."""
        out = {}
        for output in self._output[type].values():
            out[output.id] = output.state
        return out



"""Helper dir for BoneIO."""
from typing import Callable, TypeVar, Any

from boneio.helper.gpio import (
    edge_detect,
    setup_input,
    setup_output,
    read_input,
    write_output,
    GpioBaseClass,
    configure_pin,
)
from boneio.helper.oled import make_font
from boneio.helper.mqtt import BasicMqtt
from boneio.helper.stats import HostData, host_stats
from boneio.helper.yaml_util import (
    CustomValidator,
    load_yaml_file,
    schema_file,
    load_config_from_string,
    load_config_from_file,
)
from boneio.helper.state_manager import StateManager
from boneio.helper.ha_discovery import (
    ha_light_availabilty_message,
    ha_switch_availabilty_message,
    ha_sensor_temp_availabilty_message,
    ha_button_availabilty_message,
    ha_binary_sensor_availabilty_message,
    ha_input_availabilty_message,
    ha_sensor_availabilty_message,
    ha_adc_sensor_availabilty_message,
)
from boneio.helper.exceptions import GPIOInputException, I2CError
from boneio.helper.queue import UniqueQueue
from boneio.helper.timeperiod import TimePeriod

CALLABLE_T = TypeVar("CALLABLE_T", bound=Callable[..., Any])
CALLBACK_TYPE = Callable[[], None]


def callback(func: CALLABLE_T) -> CALLABLE_T:
    """Annotation to mark method as safe to call from within the event loop."""
    setattr(func, "_boneio_callback", True)
    return func


def is_callback(func: Callable[..., Any]) -> bool:
    """Check if function is safe to be called in the event loop."""
    return getattr(func, "_boneio_callback", False) is True


__all__ = [
    "CustomValidator",
    "load_yaml_file",
    "HostData",
    "host_stats",
    "setup_input",
    "setup_output",
    "edge_detect",
    "read_input",
    "write_output",
    "make_font",
    "ha_light_availabilty_message",
    "ha_switch_availabilty_message",
    "ha_sensor_availabilty_message",
    "ha_adc_sensor_availabilty_message",
    "ha_sensor_temp_availabilty_message",
    "ha_binary_sensor_availabilty_message",
    "ha_button_availabilty_message",
    "ha_input_availabilty_message",
    "GPIOInputException",
    "I2CError",
    "GpioBaseClass",
    "StateManager",
    "configure_pin",
    "BasicMqtt",
    "UniqueQueue",
    "schema_file",
    "load_config_from_string",
    "load_config_from_file",
    "TimePeriod",
]



from distutils.command.config import config
import logging
import asyncio
from datetime import datetime

from struct import unpack
import json
import os

from boneio.const import (
    ADDRESS,
    MODEL,
    SENSOR,
    STATE,
    ONLINE,
    OFFLINE,
    BASE,
    LENGTH,
    REGISTERS,
)
from boneio.helper import BasicMqtt
from boneio.helper.ha_discovery import modbus_sensor_availabilty_message
from boneio.helper.timeperiod import TimePeriod
from boneio.helper.config import ConfigHelper
from boneio.modbus import Modbus


_LOGGER = logging.getLogger(__name__)


def float32(result, base, addr):
    """Read Float value from register."""
    low = result.getRegister(addr - base)
    high = result.getRegister(addr - base + 1)
    data = bytearray(4)
    data[0] = high & 0xFF
    data[1] = high >> 8
    data[2] = low & 0xFF
    data[3] = low >> 8
    val = unpack("f", bytes(data))
    return val[0]


def floatsofar(result, base, addr):
    """Read Float value from register."""
    low = result.getRegister(addr - base)
    high = result.getRegister(addr - base + 1)
    return high + low


def multiply0_1(result, base, addr):
    low = result.getRegister(addr - base)
    return round(low * 0.1, 4)


def multiply0_01(result, base, addr):
    low = result.getRegister(addr - base)
    return round(low * 0.01, 4)


def multiply10(result, base, addr):
    low = result.getRegister(addr - base)
    return round(low * 10, 4)


def regular_result(result, base, addr):
    return result.getRegister(addr - base)


CONVERT_METHODS = {
    "float32": float32,
    "multiply0_1": multiply0_1,
    "multiply0_01": multiply0_01,
    "floatsofar": floatsofar,
    "multiply10": multiply10,
    "regular": regular_result,
}
REGISTERS_BASE = "registers_base"


def open_json(model: str) -> dict:
    """Open json file."""
    file = f"{os.path.join(os.path.dirname(__file__))}/{model}.json"
    with open(file, "r") as db_file:
        datastore = json.load(db_file)
        return datastore


class ModbusSensor(BasicMqtt):
    """Represent Modbus sensor in BoneIO."""

    SensorClass = None
    DefaultName = "ModbusSensor"

    def __init__(
        self,
        modbus: Modbus,
        address: str,
        model: str,
        config_helper: ConfigHelper,
        id: str = DefaultName,
        update_interval: TimePeriod = TimePeriod(seconds=60),
        **kwargs,
    ):
        """Initialize Modbus sensor class."""
        super().__init__(
            id=id or address,
            topic_type=SENSOR,
            topic_prefix=config_helper.topic_prefix,
            **kwargs,
        )
        self._config_helper = config_helper
        self._modbus = modbus
        self._db = open_json(model=model)
        self._model = self._db[MODEL]
        self._address = address
        self._discovery_sent = False
        self._update_interval = update_interval

    def _send_ha_autodiscovery(
        self, id: str, sdm_name: str, sensor_id: str, **kwargs
    ) -> None:
        """Send HA autodiscovery information for each Modbus sensor."""
        _LOGGER.debug("Sending HA discovery for sensor %s %s.", sdm_name, sensor_id)
        self._send_message(
            topic=(
                f"{self._config_helper.ha_discovery_prefix}/{SENSOR}/{self._config_helper.topic_prefix}{id}"
                f"/{id}{sensor_id.replace('_', '').replace(' ', '').lower()}/config"
            ),
            payload=modbus_sensor_availabilty_message(
                topic=self._config_helper.topic_prefix,
                id=id,
                name=sdm_name,
                model=self._model,
                sensor_id=sensor_id,
                **kwargs,
            ),
        )

    def _send_discovery_for_all_registers(self, register: int = 0) -> bool:
        """Send discovery message to HA for each register."""
        if register > 0:
            for data in self._db[REGISTERS_BASE]:
                for register in data[REGISTERS]:
                    value_template = (
                        f'{{{{ value_json.{register.get("name").replace(" ", "")} | '
                        f'{register.get("ha_filter", "round(2)")} }}}}'
                    )
                    kwargs = {
                        "unit_of_measurement": register.get("unit_of_measurement"),
                        "state_class": register.get("state_class"),
                        "value_template": value_template,
                        "sensor_id": register.get("name"),
                    }
                    device_class = register.get("device_class")
                    if device_class:
                        kwargs["device_class"] = device_class
                    self._send_ha_autodiscovery(
                        id=self._id,
                        sdm_name=self._name,
                        state_topic_base=data[BASE],
                        **kwargs,
                    )
            return datetime.now()
        return False

    async def check_availability(self) -> None:
        """Get first register and check if it's available."""
        if (
            not self._discovery_sent
            or (datetime.now() - self._discovery_sent).seconds > 3600
        ) and self._config_helper.topic_prefix:
            self._discovery_sent = False
            first_register_base = self._db[REGISTERS_BASE][0]
            register_method = first_register_base.get("register_type", "input")
            # Let's try fetch register 2 times in case something wrong with initial packet.
            for _ in [0, 1]:
                register = await self._modbus.read_single_register(
                    unit=self._address,
                    address=first_register_base[REGISTERS][0][ADDRESS],
                    method=register_method,
                )
                if register:
                    self._discovery_sent = self._send_discovery_for_all_registers(
                        register
                    )
                    await asyncio.sleep(2)
                    break

    async def send_state(self) -> None:
        """Fetch state periodically and send to MQTT."""
        update_interval = self._update_interval.total_seconds
        payload_online = OFFLINE
        while True:
            await self.check_availability()
            for data in self._db[REGISTERS_BASE]:
                values = await self._modbus.read_multiple_registers(
                    unit=self._address,
                    address=data[BASE],
                    count=data[LENGTH],
                    method=data.get("register_type", "input"),
                )
                if payload_online == OFFLINE and values:
                    _LOGGER.info("Sending online payload about device.")
                    payload_online = ONLINE
                    self._send_message(
                        topic=f"{self._config_helper.topic_prefix}/{self._id}{STATE}",
                        payload=payload_online,
                    )
                if not values:
                    if update_interval < 600:
                        # Let's wait litte more for device.
                        update_interval = update_interval * 1.5
                    else:
                        # Let's assume device is offline.
                        payload_online = OFFLINE
                        self._send_message(
                            topic=f"{self._config_helper.topic_prefix}/{self._id}{STATE}",
                            payload=payload_online,
                        )
                    _LOGGER.warn(
                        "Can't fetch data from modbus device. Will sleep for %s seconds",
                        update_interval,
                    )
                    self._discovery_sent = False
                    continue
                elif update_interval != self._update_interval.total_seconds:
                    update_interval = self._update_interval.total_seconds
                output = {}
                for register in data["registers"]:
                    output[register.get("name").replace(" ", "")] = CONVERT_METHODS[
                        register.get("return_type", "regular")
                    ](result=values, base=data[BASE], addr=register.get("address"))
                self._send_message(
                    topic=f"{self._send_topic}/{data[BASE]}",
                    payload=output,
                )
            await asyncio.sleep(update_interval)



from __future__ import annotations

import asyncio
import logging
from boneio.helper import configure_pin
from boneio.const import ID, RX, TX, UART, REGISTERS
from pymodbus.client.sync import (
    BaseModbusClient,
    ModbusSerialClient,
)
from pymodbus.register_read_message import ReadInputRegistersResponse
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.constants import Endian
from typing import Any
from pymodbus.exceptions import ModbusException
from pymodbus.pdu import ModbusResponse

_LOGGER = logging.getLogger(__name__)


class Modbus:
    """Represent modbus connection over chosen UART."""

    def __init__(self, uart: dict[str, Any]) -> None:
        """Initialize the Modbus hub."""

        configure_pin(pin=uart[RX], mode=UART)
        configure_pin(pin=uart[TX], mode=UART)
        self._uart = uart

        # generic configuration
        self._client: BaseModbusClient | None = None
        self._lock = asyncio.Lock()
        try:
            self._client = ModbusSerialClient(
                port=self._uart[ID],
                method="rtu",
                baudrate=9600,
                stopbits=1,
                bytesize=8,
                parity="N",
            )
            self._read_methods = {
                "input": self._client.read_input_registers,
                "holding": self._client.read_holding_registers,
            }
        except ModbusException as exception_error:
            _LOGGER.error(exception_error)

    async def async_close(self) -> None:
        """Disconnect client."""
        async with self._lock:
            if self._client:
                try:
                    self._client.close()
                except ModbusException as exception_error:
                    _LOGGER.error(exception_error)
                del self._client
                self._client = None
                _LOGGER.warning("modbus communication closed")

    def _pymodbus_connect(self) -> bool:
        """Connect client."""
        try:
            return self._client.connect()  # type: ignore[union-attr]
        except ModbusException as exception_error:
            _LOGGER.error(exception_error)
            return False

    async def read_single_register(
        self, unit: int, address: int, count: int = 2, method: str = "input"
    ) -> float | None:
        """Call sync. pymodbus."""
        async with self._lock:
            if not self._pymodbus_connect():
                _LOGGER.error("Can't connect to Modbus.")
                return None
            kwargs = {"unit": unit, "count": count} if unit else {}
            try:
                result: ReadInputRegistersResponse = self._read_methods[method](
                    address, **kwargs
                )
            except ModbusException as exception_error:
                _LOGGER.error(exception_error)
                return None
            if not hasattr(result, REGISTERS):
                _LOGGER.error(str(result))
                return None
            return BinaryPayloadDecoder.fromRegisters(
                result.registers, byteorder=Endian.Big, wordorder=Endian.Big
            ).decode_32bit_float()

    async def read_multiple_registers(
        self, unit: int, address: int, count: int = 2, method: str = "input"
    ) -> ModbusResponse:
        """Call sync. pymodbus."""
        async with self._lock:
            if not self._pymodbus_connect():
                _LOGGER.error("Can't connect to Modbus.")
                return None
            kwargs = {"unit": unit, "count": count} if unit else {}
            try:
                result: ReadInputRegistersResponse = self._read_methods[method](
                    address, **kwargs
                )
            except ModbusException as exception_error:
                _LOGGER.error(exception_error)
                return None
            if not hasattr(result, REGISTERS):
                _LOGGER.error(str(result))
                return None
            return result



"""
Module to provide basic config options.
"""
from boneio.const import BONEIO, HOMEASSISTANT


class ConfigHelper:
    def __init__(
        self,
        topic_prefix: str = BONEIO,
        ha_discovery: bool = True,
        ha_discovery_prefix: str = HOMEASSISTANT,
    ):
        self._topic_prefix = topic_prefix
        self._ha_discovery = ha_discovery
        self._ha_discovery_prefix = ha_discovery_prefix

    @property
    def topic_prefix(self) -> str:
        return self._topic_prefix

    @property
    def ha_discovery(self) -> bool:
        return self._ha_discovery

    @property
    def ha_discovery_prefix(self) -> str:
        return self._ha_discovery_prefix

    @property
    def cmd_topic_prefix(self) -> str:
        return f"{self.topic_prefix}/cmd/"

    @property
    def subscribe_topic(self) -> str:
        return f"{self.cmd_topic_prefix}+/+/#"



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


with open("boneio/version.py") as f:
    exec(f.read())

setup(
    name="boneio",
    version=__version__,  # type: ignore # noqa: F821,
    description="Python App for BoneIO",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/boneIO-eu/app_bbb",
    download_url="https://github.com/boneIO-eu/app_bbb/archive/{}.zip".format(
        __version__  # type: ignore # noqa: F821
    ),
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    license="GNU General Public License v3.0",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=[
        "Adafruit-BBIO==1.2.0",
        "Adafruit-Blinka==6.20.4",
        "adafruit-circuitpython-mcp230xx==2.5.3",
        "adafruit-circuitpython-mcp9808==3.3.10",
        "adafruit-circuitpython-pct2075==1.1.12",
        "adafruit-circuitpython-register==1.9.8",
        "adafruit-circuitpython-busdevice==5.1.4",
        "Adafruit-PlatformDetect==3.19.5",
        "Adafruit-PureIO==1.1.9",
        "asyncio-mqtt==0.12.1",
        "Cerberus==1.3.4",
        "colorlog==6.6.0",
        "gpio==0.3.0",
        "luma.core==2.3.1",
        "luma.oled==3.8.1",
        "Pillow==9.0.1",
        "paho-mqtt==1.6.1",
        "psutil==5.9.0",
        "PyYAML==6.0",
        "pymodbus==2.5.3",
        "pyserial-asyncio==0.6",
        "typing_extensions==4.1.0",
    ],
    entry_points={"console_scripts": ["boneio=boneio.bonecli:main"]},
)



"""ADC GPIO BBB sensor."""
import logging
import asyncio

from boneio.const import SENSOR
from boneio.helper import BasicMqtt
from boneio.helper.timeperiod import TimePeriod

try:
    import Adafruit_BBIO.ADC as ADC
except ModuleNotFoundError:

    class ADC:
        def __init__(self):
            pass

    pass

_LOGGER = logging.getLogger(__name__)


def initialize_adc():
    ADC.setup()


class GpioADCSensor(BasicMqtt):
    """Represent Gpio ADC sensor."""

    def __init__(
        self, pin: str, update_interval: int = TimePeriod(seconds=60), **kwargs
    ) -> None:
        """Setup GPIO ADC Sensor"""
        super().__init__(topic_type=SENSOR, **kwargs)
        self._pin = pin
        self._update_interval = update_interval
        _LOGGER.debug("Configured sensor pin %s", self._pin)

    @property
    def state(self):
        """Give rounded value of temperature."""
        return round(ADC.read(self._pin) * 1.8, 2)

    async def send_state(self):
        """Fetch temperature periodically and send to MQTT."""
        while True:
            self._send_message(
                topic=self._send_topic,
                payload=self.state,
            )
            await asyncio.sleep(self._update_interval.total_seconds)
